#!/usr/bin/ruby

$:.unshift File.join(File.dirname(__FILE__),'.')
require 'getoptlong'
require 'tempfile'
require 'vcf_variant.rb'
require 'sample.rb'
require 'fileutils'

#This class reads in per sample vcf files generated from Atlas-SNP, Atlas-Indel and  out multi-sample VCF file

class VcfPrinter
	def initialize(vcf_files, out_file)
		@samples = Hash.new
		@variants = Hash.new
		vcf_files.each do |vcf_file|
			if File.exist?(vcf_file) and File.fnmatch("*.vcf",vcf_file)
				@samples[File.basename(vcf_file, '.vcf')]=vcf_file
			else
				raise "Either #{vcf_file} does not exist or the file does not have proper .vcf extension"
			end
		end
		puts "Number of Samples: #{@samples.length}"	
		@samples.values.each do |fileName|
			lines = File.open(fileName, 'r')
			lines.each do |line|
				if line =~ /^#/
					next
				else
					var = VcfLine.new(line)
					if @variants.keys.include? var.uniqPos
						next
					else 
						@variants[var.uniqPos]=var
					end
				end
			end
			lines.close()
		end
		puts "Number of Variants: #{@variants.length}"
		File.open(out_file,"w") do |writer|
		#Printing VCF header information
			writer.puts "##fileformat=VCFv4.0"
			time = Time.new
			writer.puts "##fileDate=#{time.ctime}"
			writer.puts "##INFO=<ID=NS,Number=1,Type=Integer,Description=\"Number of Samples With Data\">"
			writer.puts "##INFO=<ID=DP,Number=1,Type=Integer,Description=\"Total Depth\">"
			writer.puts "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">"
			writer.puts "##FORMAT=<ID=VR,Number=1,Type=Integer,Description=\"Variant Read Depth\">"
			writer.puts "##FORMAT=<ID=RR,Number=1,Type=Integer,Description=\"Reference Read Depth\">"
			writer.puts "##FORMAT=<ID=DP,Number=1,Type=Integer,Description=\"Total Read Depth\">"
			writer.puts "##FORMAT=<ID=GQ,Number=1,Type=Integer,Description=\"Genotype Quality\">"
			writer.puts "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t#{@samples.keys.join("\t")}"
		#Printing column names	
			@variants.values.each do |variant|
				writer.puts "#{variant.chr}\t#{variant.coor}\t#{variant.id}\t#{variant.refBase}\t#{variant.altBase}\t#{variant.qual}\t#{variant.filter}\t.\t#{variant.format}"
			end
		end
		commandLineSort(out_file)
	end

	def print(outfile, bamFiles, reference)
	#for each sample loop through variants at the end of each sample add a column to the outfile
		bams = Hash.new
		bamFiles.each do |bam|
			if File.exist?(bam) and File.fnmatch("*.bam",bam)
                                bams[File.basename(bam, '.bam')]=bam
                        else
                                raise "Either #{bam} does not exist or the file does not have proper .bam extension"
                        end

		end
		bedFile = createBedFile(@variants.values)
		@samples.each do |sampleName, vcfFile|
			bamFile = bams[sampleName]
			sample = Sample.new(sampleName, vcfFile, bamFile, bedFile, reference)
			tmpfile = Tempfile.new('tmp')
			File.open(outfile,"r").each_line do |line|
				if line =~ /^#/
					tmpfile.puts(line)
                	        else	
					var = VcfLine.new(line)
					if sample.vcf.vcf.keys.include? var.pos
						tmpfile.puts("#{line.strip}\t#{sample.vcf.vcf[var.pos].genotypeInfo}")
					elsif sample.pileup.pileup.keys.include? var.pos
						tmpfile.puts("#{line.strip}\t#{sample.pileup.pileup[var.pos].genotypeInfo}")
					else
						tmpfile.puts("#{line.strip}\t./.:.:.:.:.")
					end
				end
			end
			tmpfile.close()
			FileUtils.mv(tmpfile.path, outfile)
		end
	end

	#creates INFO column in the final multi-sample VCF
	def generateInfoField(outFile)
		tFile = Tempfile.new('tmp')
		puts tFile.path
		File.open(outFile,"r").each_line do |line|
                	if line =~ /^#/
                        	tFile.puts(line)
                        else
				totalDepth = 0
		                totalSample = 0
				line = line.strip.split("\t")
				(9..line.length-1).each do |i|
					depth = line[i].split(':')[3].to_i
					if line[i] == './.:.:.:.:.'
						next
					else
						totalSample += 1
						if depth == '.'
							next
						else
							totalDepth += depth
						end
					end
				end
				infoField = "NS=#{totalSample};DP=#{totalDepth}"
				tFile.puts("#{line[0..6].join("\t")}\t#{infoField}\t#{line[8..line.length-1].join("\t")}")
			end
		end
		tFile.close()
		FileUtils.mv(tFile.path, outFile)
	end


	# collapses altBase
	def collapseVariants(outFile)
		tfile = Tempfile.new('tmp')
		file = File.open(outFile, 'r')
		prevLine = ''
		while (currentLine=file.gets)
			if currentLine =~ /^#/
				tfile.puts(currentLine)
			else
				if prevLine == ''
					prevLine = currentLine
				else
					prevObj = VcfLine.new(prevLine)
					currentObj = VcfLine.new(currentLine)
					if prevObj.chr == currentObj.chr and prevObj.coor == currentObj.coor and prevObj.refBase == currentObj.refBase
						pline = "#{prevObj.chr}\t#{prevObj.coor}\t#{prevObj.id}\t#{prevObj.refBase}\t#{prevObj.altBase},#{currentObj.altBase}\t#{prevLine.split("\t")[5..prevLine.split("\t").length].join("\t")}"
						prevLine = pline
					else
						tfile.puts(prevLine)
						prevLine = currentLine
					end
				end
			end
		end
		tfile.puts(prevLine)
		tfile.close()
		FileUtils.mv(tfile.path, outFile)
	end

	# creates a bed file with list of variant positions 			
	def createBedFile(vars)
		tfile = Tempfile.new('tmp')
		puts tfile.path
		vars.each do |var|
			tfile.puts "#{var.chr}\t#{var.coor}"
		end
		tfile.close
		return tfile.path
	end

	# sorts our vcf file
	def commandLineSort(outFile)
		tfile = Tempfile.new('tmp')
		#substitute X,Y,M with 23,24,25 using sed
		`sed -i -e 's/^X/23/' -e 's/^Y/24/' -e 's/^M/25/' #{outFile}`
		#copy header
		`grep -e '^#' #{outFile} >#{tfile.path}`
		#sort rows
		`grep -v -e '^#' #{outFile} | sort -t  $'\t' -k1 -k2 -n >>#{tfile.path}`
		#revert back to X,Y,M from 23,24,25
		`sed -i -e 's/^23/X/' -e 's/^24/Y/' -e 's/^25/M/' #{tfile.path}`
		FileUtils.mv(tfile.path, outFile)
		return outFile
	end


	private :commandLineSort, :createBedFile

	def self.usage
		puts "USAGE:\n\truby vcfPrinter.rb \n\t -i \"*.vcf\" \n\t -o outputfile \n\t -b \"*.bam\" \n\t -r reference.fasta \n"
		puts "The individual sample VCF filenames must follow the convention:\n\t [sample_name].vcf \nand corresponding BAM files should be named\n\t [sample_name].bam"
	end
	

	def self.processArgs()
		if ARGV.length < 4
			puts usage()
			Process.exit(0)
		else
		opts = GetoptLong.new(
		["--vcfFiles","-i",GetoptLong::REQUIRED_ARGUMENT],
		["--out","-o",GetoptLong::REQUIRED_ARGUMENT],
		["--bams","-b",GetoptLong::REQUIRED_ARGUMENT],
		["--reference","-r",GetoptLong::REQUIRED_ARGUMENT],
		["--usage","-h",GetoptLong::NO_ARGUMENT]
		)
		options = {}
		begin
			opts.each do |opt, arg|
				case opt
					when "--vcfFiles"
						options['vcf_files'] = Dir.glob(arg) 
					when "--out"
						options['out_file'] = arg
					when "--bams"
						options['bams'] = Dir.glob(arg)
					when "--reference"
						options['reference'] = arg
					when "--usage"
				        	puts usage()
						Process.exit(0)
					end
			end
		return options
		rescue
			VcfPrinter.usage()
			Process.exit(0)
		end
		end	
	end

end

if __FILE__ == $0
	args = VcfPrinter.processArgs()
	vcfPrintObj = VcfPrinter.new(args['vcf_files'], args['out_file'])
	vcfPrintObj.print(args['out_file'], args['bams'], args['reference'])
	vcfPrintObj.generateInfoField(args['out_file'])
	vcfPrintObj.collapseVariants(args['out_file'])
end
